/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include "pim/mail/MailSyncSource.h"
#include "spds/SyncItemStatus.h"
#include "mail/MailMessage.h"
#include "spds/constants.h"
#include "base/util/utils.h"
#include "base/util/StringBuffer.h"
#include "pim/utils.h"
#include "spds/FolderData.h"
#include "notify/timed_msgbox.h"
#include "mail/MailSyncSourceConfig.h"
#include "mail/MailAccount.h"

#include "HwndFunctions.h"
#include "OutOfMemoryException.h"
#include "pim/mail/MessagingManager.h"

#include "pim/mail/MessagingStatus.h" // ItemData

#define ASK_REFRESH_MESSAGE TEXT("A sync to replace your local email with your mail server email has been requested. All the local mails will be deleted before adding the server items. Are you sure you want to continue?")
#define ASK_REFRESH_TIMEOUT 10

MailSyncSource* MailSyncSource::pinstance = NULL;

MailSyncSource* MailSyncSource::getInstance(const wchar_t* name, MailSyncSourceConfig *sc)
{
	if (pinstance == NULL) {
		pinstance = new MailSyncSource(name, sc); 
	}

	return pinstance;
}

MailSyncSource* getMailSyncSource(const wchar_t* name, MailSyncSourceConfig *sc)
{
	return MailSyncSource::getInstance(name, sc);
}

/**
 * Returns a (new allocated) SyncItem* from a MailClientData*
 */
SyncItem* MailSyncSource::buildSyncItem(MailClientData* m)
{
    MailMessage& msg = m->getEmailData()->getEmailItem();
    WString msgId;
    StringBuffer parentRole = m->getParentRole();
    const wchar_t* emailSyncMlType = TEXT("application/vnd.omads-email+xml");

    msgId = m->getEntryId();
     
    if (msgId.empty()) {
        LOG.error("error getting message ID");

        return syncItem;
    }
    
    SyncItem *syncItem = NULL;
    WString idMessage(TEXT(""));

    if (!serverHasMeaCap) { // without MEA we don't have account ID on server        
        StringBuffer folderRole = m->getParentRole();        
        if (folderRole == ROLE_INBOX) {
            idMessage = INBOX_MESSAGE_KEY_WPREFIX;
        } else if (folderRole == ROLE_OUTBOX) {
            idMessage = OUTBOX_MESSAGE_KEY_WPREFIX; 
        }
    }
    idMessage += msgId.c_str();
    syncItem = new SyncItem(idMessage);   

    char* toSend = m->getEmailData()->format();
    syncItem->setData(toSend, strlen(toSend) * sizeof(char));
    syncItem->setDataType(emailSyncMlType);
    
    const wchar_t* sourceParent =  m->getParent();   
    syncItem->setSourceParent(sourceParent);

    if (toSend)
        delete [] toSend;

    return syncItem;
}

/**
 * returns a newly allocated SyncItem from an folder ItemData
 */
SyncItem* MailSyncSource::buildSyncItem(ItemData* itemData)
{
    SyncItem *syncItem            = NULL;
    char* syncMlFolderData        = NULL;
    const wchar_t* folderId       = NULL;
    const wchar_t* folderParentId = NULL;
    const wchar_t* folderSyncMlType = TEXT("application/vnd.omads-folder+xml");
    FolderData folderData;
    
    if (itemData == NULL) {
        LOG.error("invalid item data");

        return syncItem;
    }

    StringBuffer folderRole = itemData->getParentRole();

    folderData.setRole( folderRole );
    if (folderRole == ROLE_INBOX) {
        folderData.setName("Inbox");
    } else if (folderRole == ROLE_OUTBOX) {
        folderData.setName("Outbox");
    }

    folderData.setID( itemData->getId() );
    folderId = itemData->getId();

    if ((syncMlFolderData = folderData.format()) == NULL) {
        LOG.error("error getting folder syncML formatted data");

        return syncItem;
    }

    syncItem = new SyncItem( folderId  );
    syncItem->setData(syncMlFolderData, strlen(syncMlFolderData) * sizeof(char));
    syncItem->setDataType(folderSyncMlType);

    if (!serverHasMeaCap) { // without MEA we don't have account ID on server
        syncItem->setSourceParent(SERVER_ROOT_FOLDER_W);
    } else {
        const wchar_t* wParentId = itemData->getParentId(); 
        const char* parentId     = NULL;

        if (wParentId) {
            parentId = toMultibyte( wParentId );
            folderData.setParent( parentId );
            folderParentId = itemData->getParentId();
            delete [] parentId;
        } else {
            LOG.error("error getting parent ID from item");

            return syncItem;
        }

        syncItem->setSourceParent(folderParentId);
    }

    if (syncMlFolderData) {
        delete [] syncMlFolderData;
    }

    return syncItem;
}

/**
 * returns a newly allocated SyncItem from a mail account ID
 */
SyncItem* MailSyncSource::buildSyncItem(const wchar_t* accountId)
{
    SyncItem *syncItem  = NULL;
    MailAccount* accountData = messagingManager.getAccountById(accountId);

    if (accountData == NULL) {
        LOG.error("can't get account with ID %S", accountId);
        return syncItem;
    }
    
    accountData->setRole(ROLE_ACCOUNT);

    const wchar_t* syncMlType = TEXT("application/vnd.omads-folder+xml");
    char* syncMlAccountData        = NULL;

    if ((syncMlAccountData = accountData->format()) == NULL) {
        LOG.error("error getting syncML Mail Account formatted data");

        return syncItem;
    }

    syncItem = new SyncItem( accountId  );
    syncItem->setData(syncMlAccountData, strlen(syncMlAccountData) * sizeof(char));
    syncItem->setDataType(syncMlType);
    syncItem->setSourceParent(SERVER_ROOT_FOLDER_W); // all accounts have "/" as parent

    if (syncMlAccountData) {
        delete [] syncMlAccountData;
    }

    return syncItem;
}

int MailSyncSource::parseMailData(SyncItem &item, MailClientData &m, bool isUpdate)
{
    int errorCode;

    char *msg = new char[item.getDataSize() + 1];
    memset(msg, 0, item.getDataSize() + 1);
    memcpy(msg, item.getData(), item.getDataSize());

    errorCode = m.getEmailData()->parse(msg);

    //
    // -2 is a real error
    // -1 is an error only if it is a new item but not if it is an update.
    //
    if (errorCode == 0 || (errorCode == -1 && isUpdate)) {
        // nothing to do.
        // it continue usually
    } else {
        delete [] msg;
        return MAIL_ITEM_PARSE_ERROR;
    }

    // Set the parent folder's key + role
    wstring parentKey;
    StringBuffer parentRole;
    wstring targetParent(item.getTargetParent());
    if (targetParent.empty() || targetParent.find(TEXT("/")) != wstring::npos) {
        //
        // Backward compatibility: manage old keys like "I/<key>"
        //
        parentKey = getParentFolderFromItemKey(item.getKey());
        wstring letter;
        if (parentKey.empty()) {
            delete [] msg;
            return MAIL_ITEM_STORE_ERROR;
        }

        letter = item.getKey()[0];
        parentRole = messagingManager.getFolderRoleFromLetter(letter);
    } 
    else {
        parentKey = item.getTargetParent();
        parentRole = messagingManager.getFolderRole(parentKey);
    }
    m.setParent(parentKey.c_str());
    m.setParentRole(parentRole);

    // Set other params to handle the inclusive filter (TODO: get from MSSConfig!)
    m.setMaxMailMessageSize(getMaxMailMessageSize());
    m.setCurrentMessageSizeFilter(((MailSyncSourceConfig&)getConfig()).getAttachSize());
    m.setIsSyncInclusive(getIsSyncInclusive());
    m.setIncludeAttachment(getIncludeAttachment());

    delete [] msg;
    return MAIL_ITEM_NO_ERROR;
}

static bool setFolderData(SyncItem &item, /*const wchar_t *parent*/ FolderData &folder)
{
    int errorCode;

    char *msg = new char[item.getDataSize() + 1];
    memset(msg, 0, item.getDataSize() + 1);
    memcpy(msg, item.getData(), item.getDataSize());

    errorCode = folder.parse(msg);

    if (errorCode != 0) {
        LOG.error("Error parsing message!");
        return false;
    }

    delete [] msg;
    return true;
}



MailSyncSource::MailSyncSource(const wchar_t* name, MailSyncSourceConfig *sc) : 
                SyncSource(name, sc),
                messagingManager(*sc),
                isMailInOutbox(false),
                isMailInInbox(false),
                failedSendMailInOutbox(false)
{
    path = wstrdup(TEXT("\\"));
    syncItem = NULL;
    maxMailMessageSize = 0;
    isSyncInclusive = false;

    // register to sync source notifications for
    // server begin/end  
    syncSourceListener  = new SyncSourceListenerServer(MAIL_SYNC_SOURCE_LISTENER_NAME);
    setSyncSourceListener(syncSourceListener);

    // check if server support MEA
    serverHasMeaCap = getRegConfig()->getServerConfig().getMultipleEmailAccount();
    if (serverHasMeaCap) {
        LOG.info("server has MEA support");
    } else {
        LOG.info("no MEA support from server: going backward compatibility mode");
    }
}

MailSyncSource::MailSyncSource(MailSyncSource& s) : SyncSource(s), 
                               messagingManager((MailSyncSourceConfig&)(s.getConfig())),
                               isMailInOutbox(false),
                               isMailInInbox(false),
                               failedSendMailInOutbox(false)
{
    path = NULL;
    setLastSync(s.getLastSync());
    setNextSync(s.getNextSync());
    setMaxMailMessageSize(s.getMaxMailMessageSize());
    setLastAnchor(s.getLastAnchor());
    setNextAnchor(s.getNextAnchor());
    //setErrorHandler(s.getErrorHandler());
    setFilter(s.getFilter());
    setPath(s.path);
}

MailSyncSource::~MailSyncSource() {
    if (syncItem) {
        syncItem = NULL;
    }
    if (path) {
        delete [] path; path = NULL;
    }

    if (syncSourceListener) {
        unsetSyncSourceListener(MAIL_SYNC_SOURCE_LISTENER_NAME);
    }
}

void MailSyncSource::setPath(const wchar_t* p) {
    if (path)
        delete [] path;

    if (p == NULL)
        path = wstrdup(TEXT(""));
    else
        path = wstrdup(p);
}

const wchar_t* MailSyncSource::getPath() {
    return path;
}


void MailSyncSource::setIsMailInOutbox(bool value){
    isMailInOutbox = value;

}

bool MailSyncSource::getIsMailInOutbox() {
    return isMailInOutbox;
}

void MailSyncSource::setIsMailInInbox(bool value){
    isMailInInbox = value;
}

bool MailSyncSource::getIsMailInInbox() {
    return isMailInInbox;
}

void MailSyncSource::setFailedSendMailInOutbox(bool value){
    failedSendMailInOutbox = value;
}

bool MailSyncSource::getFailedSendMailInOutbox() {
    return failedSendMailInOutbox;
}


/*
 * Return the first SyncItem of all. It is used in case of slow
 * sync and retrieve the entire data source content.
 */
SyncItem* MailSyncSource::getFirstItem() {

    LOG.debug("getFirstItem");
    iAll = 0;
    return getNextItem();
}

/*
* Return the next SyncItem of all. It is used in case of slow or refresh sync
* and retrieve the entire data source content.
*/
SyncItem* MailSyncSource::getNextItem() {
    
    syncItem = NULL;
    
    LOG.debug("getNextItem Item NULL size of iAll: %d allItem: %d", iAll, allItems.size());
        
    if (iAll >= allItems.size()) {
        return NULL;
    }

    // Get the next item of all
    ItemData* itemData = (ItemData*)allItems[iAll];
    unsigned long itemType = 0L;

    if (!itemData) {
        LOG.debug("getNextItem Item NULL");
        return NULL;
    }

    itemType = itemData->getType();

    if (itemType == MAPI_STORE) {
        const wchar_t* accountId = itemData->getId();
        if ((syncItem = buildSyncItem(accountId)) == NULL) {
            LOG.error("error creating sync item for account ID %S", accountId);
        } else {
            syncItem->setState(SYNC_STATE_NONE);
        }
    } else if (itemType == MAPI_FOLDER) {
        syncItem = buildSyncItem(itemData);
        syncItem->setState(SYNC_STATE_NONE);
    } else if (itemType == MAPI_MESSAGE) {
        // Fills the MailClientData object with data
        MailClientData* mail = messagingManager.completeObject(*itemData, true);    // true: like an updateItem, to send only headers
        if (mail) {
            LOG.info("getNextItem, subject: -%s-", mail->getEmailData()->getEmailItem().getSubject());
            syncItem = buildSyncItem(mail);
            syncItem->setState(SYNC_STATE_NONE);
            delete mail; mail = NULL;
        }
    } else {
        LOG.error("invalid MAPI item type: can't build sync item");
    }

    iAll ++;
    return syncItem;
}


/*
* This method is always called by the sync manager. So the setModifiedItemsPOOM is always called
*/
SyncItem* MailSyncSource::getFirstNewItem() {

    iNew = 0;
    return getNextNewItem();
}

SyncItem* MailSyncSource::getNextNewItem() {
    
    syncItem = NULL;
    if (iNew >= newItems.size()) {
        return NULL;
    }

    // Get the next item of NEW
    ItemData* itemData = (ItemData*)newItems[iNew];
    if (!itemData) {
        return NULL;
    }

    // Fills the MailClientData object with data
    MailClientData* mail = messagingManager.completeObject(*itemData, false);
    if (mail) {
        LOG.info("getNextNewItem, subject: -%s-", mail->getEmailData()->getEmailItem().getSubject());
        syncItem = buildSyncItem(mail);
        syncItem->setState(SYNC_STATE_NEW);
        delete mail; mail = NULL;
    }

    iNew ++;
    return syncItem;
}

SyncItem* MailSyncSource::getFirstUpdatedItem() {

    iMod = 0;
    return getNextUpdatedItem();
}

SyncItem* MailSyncSource::getNextUpdatedItem() {
    
    syncItem = NULL;
    if (iMod >= modItems.size()) {
        return NULL;
    }

    // Get the next item of MOD
    ItemData* itemData = (ItemData*)modItems[iMod];
    if (!itemData) {
        return NULL;
    }

    // Fills the MailClientData object with data
    MailClientData* mail = messagingManager.completeObject(*itemData, true);
    if (mail) {
        LOG.info("getNextModItem, subject: -%s-", mail->getEmailData()->getEmailItem().getSubject());
        syncItem = buildSyncItem(mail);
        syncItem->setState(SYNC_STATE_UPDATED);
        delete mail; mail = NULL;
    }

    iMod ++;
    return syncItem;
}

SyncItem* MailSyncSource::getFirstDeletedItem() {
    
    iDel = 0;
    return getNextDeletedItem();
}

SyncItem* MailSyncSource::getNextDeletedItem() {
    
    syncItem = NULL;
    if (iDel >= delItems.size()) {
        return NULL;
    }

    // Get the next item of DEL
    ItemData* itemData = (ItemData*)delItems[iDel];
    if (!itemData) {
        return NULL;
    }
    WString idMessage;
    if (!serverHasMeaCap) { // without MEA we don't have account ID on server        
        StringBuffer folderRole = itemData->getParentRole();        
        if (folderRole == ROLE_INBOX) {
            idMessage = INBOX_MESSAGE_KEY_WPREFIX;
        } else if (folderRole == ROLE_OUTBOX) {
            idMessage = OUTBOX_MESSAGE_KEY_WPREFIX;
        }                
    } 
    idMessage += itemData->getId();

    // No need to fill item's data for deleted
    LOG.info("getNextDeletedItem, ID : -%S-", idMessage);
    syncItem = new SyncItem(idMessage);
    syncItem->setState(SYNC_STATE_DELETED);
    syncItem->setSourceParent(itemData->getParentId());

    iDel ++;
    return syncItem;
}


void MailSyncSource::setItemStatus(const WCHAR* key, int status) {

    bool inOutbox = false;
    WCHAR* itemKey = (WCHAR*)key;

    LOG.debug("Item status - key: %ls, status: %i", key, status);

    if ((wcslen(key) > 2) && ((wcsstr(key, OUTBOX_MESSAGE_KEY_WPREFIX) != NULL) ||
        (wcsstr(key, INBOX_MESSAGE_KEY_WPREFIX) != NULL))  ) {
        // mangle old style key (backward compatibility < v.8.5)
        itemKey = (WCHAR*)(key + 2);
        if (wcsstr(key, OUTBOX_MESSAGE_KEY_WPREFIX) != NULL) {
            inOutbox = true;
        }
    } else {
        
        unsigned long itemType = messagingManager.getObjectType(itemKey);

        switch (itemType) {
            case MAPI_MESSAGE: {
                wstring folderID = messagingManager.getParentFolderID(itemKey);
                StringBuffer folderRole = messagingManager.getFolderRole(folderID);
                if (folderRole == ROLE_OUTBOX) {
                    inOutbox = true;
                }
                break;
            }

            case MAPI_FOLDER:
                LOG.debug("received status notification for folder item");
                break;
            case MAPI_STORE:
                LOG.debug("received status notification for account item");
                break;
            default:
                LOG.error("error getting type for item with ID %S", itemKey);
                break;
        }  
    }
    
    if (status == 500 || status == 418 || status == 419) {
        //
        // Failed to send email
        //
        SyncSourceReport* rep = getReport();
        int res = rep->getItemReportFailedCount(SERVER, COMMAND_ADD);
        BOOL found = FALSE;

        for (int i = 0; i < res; i++) {
            ItemReport* r = rep->getItemReport(SERVER, COMMAND_ADD, i);
            if (wcscmp(r->getId(), itemKey) == 0) {
                found = TRUE;
                if (r->getStatusMessage())
                    LOG.error("The mail in outbox cannot be sent. Server says: %S", r->getStatusMessage());
            }
        }
        if (!found) {
            int res = rep->getItemReportFailedCount(SERVER, COMMAND_REPLACE);
            for (int i = 0; i < res; i++) {
                ItemReport* r = rep->getItemReport(SERVER, COMMAND_REPLACE, i);
                if (wcscmp(r->getId(), itemKey) == 0) {
                    if (r->getStatusMessage())
                        LOG.error("The mail in outbox cannot be sent. Server says: %S", r->getStatusMessage());
                }
            }
        }
    } 
    else if (status == 200 || status == 201) {
        if (inOutbox) {
            // OK: email sent
            if (messagingManager.moveMessageFromOutboxToSent(itemKey)) {
                LOG.error("Error moving the message %S from outbox to sent folder", itemKey);
            }
        }

        if ((msgStatusCache = MessagingStatus::getInstance()) == NULL) {
            LOG.error("error getting status message cache!");
        } else {
            // remove succesfully sent item from cache
            if (msgStatusCache->removeCacheItem(itemKey)) {
                LOG.debug("removed message item %S from cache", itemKey);
            } else {
                LOG.error("error updating item %S status in cache", itemKey);
            }
        }
    }
}


int MailSyncSource::addItem(SyncItem& item) {

    int ret = STC_COMMAND_FAILED;
    char* posFolder = NULL;
    wstring id;

    if (!isMemoryAvailable()) {
        LOG.error(LOW_DEVICE_MEMORY);
        throw OutOfMemoryException(0, -50);    //defined in startsync //return ret;
    }

    char* itemData = (char*)item.getData();
    if(!itemData) {
        LOG.error("addItem: empty data from server");
        ret = STC_COMMAND_FAILED;
        goto finally;
    }

    //
    // Check Folder tag: accounts and folders have the tag <Folder>
    // 
    posFolder = strstr(itemData, "Folder>");
    if ((posFolder != NULL) && ((posFolder - itemData) < 10)) {
        FolderData folder;
        if (!setFolderData(item, folder)) {
            ret = STC_COMMAND_FAILED;
            goto finally;
        }

        // Set the folder's parent
        StringBuffer parent;
        parent.convert(item.getTargetParent());
        if (!parent.empty()) {
            folder.setParent(parent.c_str());
        }

        const StringBuffer& role = folder.getRole();
        if (role.icmp(ROLE_ACCOUNT)) {
            //
            // This is a NEW email account
            //
            MailAccount account = folder;
            if (messagingManager.createAccount(account)) {
                ret = STC_COMMAND_FAILED;
                goto finally;
            } 
            item.setKey(account.getID());
        }
        else {
            // check if server support MEA to know if we need to 
            // discard folder creation for backward compatibility
            if (serverHasMeaCap) {
                // This is a NEW email folder
                if (messagingManager.createFolder(folder)) {
                    ret = STC_COMMAND_FAILED;
                    goto finally;
                }

                item.setKey(folder.getID());
            } else {
                StringBuffer defaultAccountId = messagingManager.getDefaultAccountID();
                wstring folderId;
                
                if (defaultAccountId.empty()) {
                    LOG.error("error getting default mail account id");
                    SyncSourceReport* report = getReport();

                    report->setState(SOURCE_ERROR);
                    report->setLastErrorCode(ERR_ITEM_ERROR);
                    report->setLastErrorMsg("can't create MAPI account");
                
                    ret = STC_COMMAND_FAILED;
                    
                    goto finally;
                }

                if (strstr(itemData, SERVER_INBOX_KEY)) {
                    folderId = messagingManager.getIdOfFolder(defaultAccountId ,ROLE_INBOX);
                } else if (strstr(itemData, SERVER_OUTBOX_KEY)) {
                    folderId = messagingManager.getIdOfFolder(defaultAccountId, ROLE_OUTBOX);
                }

                if (folderId.empty()) {
                    LOG.error("error getting inbox folder id for default account");
                } else {
                    item.setKey(folderId.c_str());
                }
            }
        }

        ret = STC_ITEM_ADDED;
        
        goto finally;
    }
    else {
        //
        // This is a NEW email message
        // Parse the data into the MailClientData object.
        //
        MailClientData m(TEXT(""));
        int itemStatus;
        if ((itemStatus = parseMailData(item, m, false)) != MAIL_ITEM_NO_ERROR) {
            if (itemStatus == MAIL_ITEM_PARSE_ERROR) {
                LOG.error("Error parsing email message");
                ret = STC_COMMAND_FAILED;
            } else { // MAIL_ITEM_STORE_ERROR are fatal for sync process
                LOG.error("error storing email message in MAPI account");
                SyncSourceReport* report = getReport();

                report->setState(SOURCE_ERROR);
                report->setLastErrorCode(ERR_ITEM_ERROR);
                report->setLastErrorMsg("can't store mail message in MAPI account");
                
                ret = STC_COMMAND_FAILED;
            }

            goto finally;
        }

        // If the message is empty, discard it
        if(m.getEmailData()->getEmailItem().empty()) {
            LOG.error("Empty email message from server");
            ret = STC_COMMAND_FAILED;

            goto finally;
        }

        // check for the storage memory...
        checkStorageMemory(0);

        // Add the mail in the messaging store. 
        // The new item's key is set inside the MailClientData object.
        int result = messagingManager.addEmail(&m);
        if (result) {
            if (result == STC_ALREADY_EXISTS) {
                ret = STC_ALREADY_EXISTS;
                LOG.debug("Mail already exists");
            }
            else {
                LOG.error("Error adding email to messaging store: code %i", result);
            }
            goto finally;
        }

        id = m.getEntryId();
        if (id.empty()) {
            LOG.error("Error in MailServer2Client::addMessage");
            goto finally;
        }

        // Ok: set the new key and the status code
        ret = STC_ITEM_ADDED;
        
        WString idMessage(TEXT(""));            
        if (!serverHasMeaCap) { // without MEA       
            StringBuffer folderRole = m.getParentRole();
            if (folderRole == ROLE_INBOX) {
                idMessage = INBOX_MESSAGE_KEY_WPREFIX;
            } else if (folderRole == ROLE_OUTBOX) {
                idMessage = OUTBOX_MESSAGE_KEY_WPREFIX;
            }            
        }
        idMessage += id.c_str();
        item.setKey(idMessage);
        

        if (HwndFunctions::wnd) {
            SendMessage(HwndFunctions::wnd, ID_MYMSG_ITEM_SYNCED_FROM_SERVER, (WPARAM)SOURCE_MAIL, NULL);
        }
        LOG.info("New email added: \"%s\"", m.getEmailData()->getEmailItem().getSubject());
    }

finally:
    return ret;
}

int MailSyncSource::updateItem(SyncItem& item) {

    int ret = STC_COMMAND_FAILED;
    wstring itemKey;

    if (!isMemoryAvailable()) {
        LOG.error(LOW_DEVICE_MEMORY);
        throw OutOfMemoryException(0, -50);    //defined in startsync
    }

    // Safe checks
    char* itemData = (char*)item.getData();    
    if(!itemData){
        LOG.error("updateItem: empty data from server");
        ret = STC_COMMAND_FAILED;
        goto finally;
    }
    itemKey = item.getKey();
    if (itemKey.empty()) {
        LOG.error("updateItem: empty item key from server");
        ret = STC_COMMAND_FAILED;
        goto finally;
    }

    //
    // Check Folder tag: accounts and folders have the tag <Folder>
    //
    char* posFolder = strstr(itemData, "Folder>");
    if ((posFolder != NULL) && ((posFolder - itemData) < 10)) {
        FolderData folder;
        if (!setFolderData(item, folder)) {
            ret = STC_COMMAND_FAILED;
            goto finally;
        }

        // Set the folder's ID
        folder.setID(item.getKey());

        // Set the folder's parent (parent is not necessary for accounts)
        StringBuffer parent;
        parent.convert(item.getTargetParent());
        folder.setParent(parent.c_str());

        const StringBuffer& role = folder.getRole();
        if (role.icmp(ROLE_ACCOUNT)) {
            //
            // Update of an email account
            //
            MailAccount account = folder;
            if (messagingManager.updateAccount(account)) {
                ret = STC_COMMAND_FAILED;
                goto finally;
            }
        }
        else {
            //
            // Update of an email folder
            //
            if (messagingManager.updateFolder(folder)) {
                ret = STC_COMMAND_FAILED;
                goto finally;
            }
        }
        ret = STC_OK;
        goto finally;
    }
    else {
        //
        // Update of an email message
        // Parse the data into the MailClientData object.
        //
        MailClientData m(item.getKey());
        int itemStatus;
        if ((itemStatus = parseMailData(item, m, true)) != MAIL_ITEM_NO_ERROR) {
            if (itemStatus == MAIL_ITEM_PARSE_ERROR) {
                LOG.error("Error parsing email message");
                ret = STC_COMMAND_FAILED;
            } else { // MAIL_ITEM_STORE_ERROR are fatal for sync process
                LOG.error("error storing email message in MAPI account");
                ret = STC_COMMAND_FAILED;
                SyncSourceReport* report = getReport();

                report->setState(SOURCE_ERROR);
                report->setLastErrorCode(ERR_ITEM_ERROR);
                report->setLastErrorMsg("can't store mail message in MAPI account");
            }

            goto finally;
        }
   
        // check for the storage memory...
        checkStorageMemory(0);
                            
        if (!serverHasMeaCap) { // without MEA   
            wstring idMessage(item.getKey());
            // Backward compatibility: old email IDs are "(O/ || I/) <key in b64>"
            if (idMessage.size() > 2 &&  (idMessage.find(INBOX_MESSAGE_KEY_WPREFIX, 0) == 0 ||
                                          idMessage.find(OUTBOX_MESSAGE_KEY_WPREFIX, 0) == 0 )) {
                idMessage = idMessage.substr(2);
                m.setEntryId(idMessage.c_str());
            }

        }
        // Update the mail in the messaging store. 
        int result = messagingManager.updateEmail(&m);
        if (result) {
            if (result == STC_ALREADY_EXISTS) {
                ret = STC_ALREADY_EXISTS;
                LOG.debug("Mail already exists");
            }
            else {
                LOG.error("Error adding email to messaging store: code %i", result);
            }
            goto finally;
        }

        // Ok: set the status code to return
        ret = STC_OK;
        if (HwndFunctions::wnd) {
            SendMessage(HwndFunctions::wnd, ID_MYMSG_ITEM_SYNCED_FROM_SERVER, (WPARAM)SOURCE_MAIL, NULL);
        }
        LOG.info("Email updated: \"%s\"", m.getEmailData()->getEmailItem().getSubject());
    }

finally:
    return ret;
}

int MailSyncSource::deleteItem(SyncItem& item) {

    int ret = STC_OK;
    const wchar_t* itemKey = item.getKey();
    
    // Backward compatibility: old email IDs are "(O/ || I/) <key in b64>"
    if ((wcslen(itemKey) > 2) && ((itemKey[0] == 'I') || (itemKey[0] == 'O')) && (itemKey[1] == '/')) {
        itemKey += 2;
    }
    const wstring parentKey = messagingManager.getParentFolderID(itemKey);
    
    if (parentKey.empty() == false) {
        StringBuffer folderRole = messagingManager.getFolderRole(parentKey);

        // discard deletes on sent items
        if (folderRole == ROLE_SENT) {
            return STC_OK;
        }
    }

    // Deletes either an account/folder/message given its ID
    if (messagingManager.deleteObject(itemKey)) {
        ret = STC_ITEM_NOT_DELETED;
    }
    else {
        if (HwndFunctions::wnd) {
            SendMessage(HwndFunctions::wnd, ID_MYMSG_ITEM_SYNCED_FROM_SERVER, (WPARAM) SOURCE_MAIL, NULL);
        }
        LOG.debug("item deleted: -%S-", item.getKey());
        return STC_OK;
    }

    return ret;
}

int MailSyncSource::beginSync() 
{
    int totalItems = 0;

    // Init messagingManager
    if (messagingManager.logon()) {
        LOG.error("Error logging on the messaging store");
        return 500;
    }

    if (getSyncMode() == SYNC_SLOW) {
        int numAccounts = messagingManager.getAccountNumber();
        //
        // SLOW SYNC: complete the list of all emails ID
        //
        allItems.clear();

        for (int i=0; i<numAccounts; i++) {
            StringBuffer accountID = messagingManager.getIdOfAccount(i);
            ArrayList mailItems;            

            if (serverHasMeaCap) {
                const wchar_t* wId = toWideChar(accountID.c_str());
                const wstring accountId(wId);
                ItemData accountItem(accountId, MAPI_STORE);
                allItems.add(static_cast<ItemData>(accountItem));
                delete [] wId;
            }
            
            // folder inbox
            wstring folderID = messagingManager.getIdOfFolder(accountID, ROLE_INBOX);

            // on slow sync send folder to server
            ItemData inboxFolderItem(folderID, MAPI_FOLDER);

            inboxFolderItem.setParentRole(ROLE_INBOX);
            if (serverHasMeaCap) {
                const wchar_t* folderParentId = toWideChar(accountID.c_str());
                const wstring parentId(folderParentId);
                inboxFolderItem.setParentId(parentId);
                delete [] folderParentId;
            
                // add folder to sync items if MEA server before the emails
                allItems.add(static_cast<ItemData>(inboxFolderItem));
            }
                        
            int ret = messagingManager.getAllMessages(mailItems, folderID, ROLE_INBOX);
            if (ret) {
                LOG.error("Error retrieving all %s mails from account #%i (code %i)", ROLE_INBOX, i+1, ret);
            }
            // set for backward compatibility the value of the ROOT/I for Inbox
            // by default.            
            if (!serverHasMeaCap) {
                for (int j = 0; j < mailItems.size(); j++) {
                    ItemData* d = (ItemData*)mailItems.get(j);
                    d->setParentId(SERVER_INBOX_PARENT_KEY_W);
                }
            }
            allItems.add(&mailItems);
            
            // in backward the inbox is sent at the end to give to the server
            // the right mapping of ROOT/I and the internad id
            // next incoming messages will have a right parentid.
            // the non mea server doesn't handle properly the inbox messages when with the 
            // inbox folder
            if (!serverHasMeaCap) {
                // Now put the inbox folder to have the right mapping for the next messages
                // of the next sync
                allItems.add(static_cast<ItemData>(inboxFolderItem));
            }

            // folder outbox
            mailItems.clear();            
            folderID = messagingManager.getIdOfFolder(accountID, ROLE_OUTBOX);
                       
            // send client outbox folder on slow syncs for mappings
            ItemData outboxFolderItem(folderID, MAPI_FOLDER);
            // ItemData outboxFolderItem(tmpFolder, MAPI_FOLDER);

            outboxFolderItem.setParentRole(ROLE_OUTBOX);
            if (serverHasMeaCap) {
                const wchar_t* folderParentId = toWideChar(accountID.c_str());
                const wstring parentId(folderParentId);
                outboxFolderItem.setParentId(parentId);
                delete [] folderParentId;
            
                // add folder to sync items if MEA server before the emails
                allItems.add(static_cast<ItemData>(outboxFolderItem));
            }
            
            ret = messagingManager.getAllMessages(mailItems, folderID, ROLE_OUTBOX);
            
            if (ret) {
                LOG.error("Error retrieving all %s mails from account #%i (code %i)", ROLE_OUTBOX, i+1, ret);
            }
            // set for backward compatibility the value of the ROOT/O for Outbox
            // by default.            
            if (!serverHasMeaCap) {
                for (int j = 0; j < mailItems.size(); j++) {
                    ItemData* d = (ItemData*)mailItems.get(j);
                    d->setParentId(SERVER_OUTBOX_PARENT_KEY_W);
                }
            }
            
            allItems.add(&mailItems);

            // in backward the outbox is sent at the end to give to the server
            // the right mapping of ROOT/O and the internad id
            // next outgoing messages will have a right parentid.
            // the non mea server doesn't handle properly the outbox messages when with the 
            // outbox folder
            if (!serverHasMeaCap) {
                allItems.add(static_cast<ItemData>(outboxFolderItem));
            }
        }

        totalItems = allItems.size();

        LOG.debug("total items: %d", totalItems);
    }
    else {
        //
        // FAST SYNC: get new/mod/del item keys from MessagingStatus
        //

        if ((msgStatusCache = MessagingStatus::getInstance()) == NULL) {
            LOG.error("error getting message status cache");
        } else {
            int numAccounts = messagingManager.getAccountNumber();
            LOG.debug("fast sync: getting mod/new/del items");
            ArrayList addedMessages;
            ArrayList updatedMessages;
            ArrayList deletedMessages;

            for (int i=0; i<numAccounts; i++) {
                int ret = 0;
                StringBuffer accountID = messagingManager.getIdOfAccount(i);
                ArrayList mailItems;

                wstring folderID = messagingManager.getIdOfFolder(accountID, ROLE_OUTBOX);

                // FIXME: add to messaging manager a method to merge a list of item (read
                // from cache) to those read from MAPI folder
                ret = messagingManager.getAllMessages(mailItems, folderID, ROLE_OUTBOX);
                if (ret) {
                    LOG.error("Error retrieving mails from folder \"%s\" account #%i (code %i)", ROLE_OUTBOX, i+1, ret);
                }
                addedMessages.add(&mailItems);
            }

            
            if (addedMessages.isEmpty() == false) {
                newItems.add(&addedMessages);
            } else {
                LOG.debug("empty cache for added messages");
            }

            updatedMessages = msgStatusCache->getMessagingStatusUpdatedItems();
            if (updatedMessages.isEmpty() == false) {
                modItems.add(&updatedMessages);
            } else {
                LOG.debug("empty cache for updated messages");
            }

            deletedMessages = msgStatusCache->getMessagingStatusDeletedItems();
            if (deletedMessages.isEmpty() == false) {
                delItems.add(&deletedMessages);
            } else {
                LOG.debug("empty cache for deleted messages");
            }

            totalItems = newItems.size() + modItems.size() + delItems.size(); 
        }
    }

    // send nr of mails from client, slow-sync, set -1 to wParam to signal that these are items from client
    if (HwndFunctions::wnd) {
        SendMessage(HwndFunctions::wnd, ID_MYMSG_TOTAL_ITEMS, SOURCE_MAIL, (LPARAM)totalItems);
    }

    return 0;
}

int MailSyncSource::endSync() {

    if ((msgStatusCache = MessagingStatus::getInstance()) == NULL) {
        LOG.error("error getting messaging status instance");
    } else {
        int checkCount = 0;
        int maxCheckCount = 20;   
        int sleepInterval = 1000; // in millisecs
        int i = 0;
        bool endCheck = false;

        // on sync end unlock messaging status cache 
        LOG.debug("unlocking message status cache");
        msgStatusCache->unlockMessagingStatusCache();
        
        Sleep(sleepInterval + 500);

        // check cache file generated by spourious events catched by advise sync
        LOG.debug("checking status cache file");
        for (i = 0; i < maxCheckCount; i++) {
            if (msgStatusCache->checkCacheFile()) {
                LOG.debug("resetting message status cache file after sync end!");
                msgStatusCache->resetCache();
                Sleep(sleepInterval);
                endCheck = false;
            } else {
                if (endCheck == false) {
                    LOG.debug("setting end flag for mail cache file");
                    endCheck = true;
                    Sleep(sleepInterval);
                    continue;       
                }  
            }

            if (endCheck) {
                break;
            } 
        }

        LOG.debug("total check number on cache file: %d", i);
    } 

    LOG.debug("MailSyncSource: endSync ended");
    
    return 0;
}

MailSyncSource& MailSyncSource::operator=(MailSyncSource& s) {
    if (this != &s) {
        setSyncMode(s.getSyncMode());
        setLastSync(s.getLastSync());
        setNextSync(s.getNextSync());
        setLastAnchor(s.getLastAnchor());
        setNextAnchor(s.getNextAnchor());
        //setErrorHandler(s.getErrorHandler());

        setFilter(s.getFilter());
        setPath(s.getPath());
        setReport(s.getReport());
        setMaxMailMessageSize(s.getMaxMailMessageSize());
    }
    return *this;
}


/**
* It deletes all the mail in the all accounts folders 
*/
int MailSyncSource::removeAllItems() {

    int ret = 0;
    MessagingManager* msgMan = getMessagingManager();
    MailSyncSourceConfig* mssc = getRegConfig()->getMailSyncSourceConfig(true);
    ArrayList al = mssc->getMailAccounts();

    for (int i = 0; i < al.size(); i++){
        MailAccount* ma = ((MailAccount*)al[i]);
        if( ma->getToBeCleaned() ){
            
            ArrayList mailitems;
            const char* tmp = toMultibyte( ma->getID() );
            wstring folderid = msgMan->getIdOfFolder(tmp, ROLE_INBOX);
            delete [] tmp; tmp = NULL;
            msgMan->getAllMessages(mailitems, folderid, ROLE_INBOX);

            for (int j = 0; j < mailitems.size(); j++){
                ItemData* msg = ((ItemData*)mailitems[j]);
                if(ret = msgMan->deleteObject(msg->getId()) != 0){
                    LOG.error("error cleaning up with error code: %d", ret);
                    return ret;
                }
            }
            mssc->setToBeCleanedFlag(ma->getName(), false);
        }
    }
    mssc->setLast(0);
    getRegConfig()->save();
    return 0;
}


wstring MailSyncSource::getParentFolderFromItemKey(const WCHAR* itemKey) {

    wstring ret;

    if (!itemKey) {
        return ret;
    }

    wstring folderLetter;
    folderLetter = itemKey[0];
    StringBuffer role = messagingManager.getFolderRoleFromLetter(folderLetter);

    StringBuffer account = messagingManager.getDefaultAccountID();
    if (account.empty()) {
        LOG.error("can't get default account ID!");
        return ret;
    }

    return messagingManager.getIdOfFolder(account, role);
}
