/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2009 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */


#include "base/util/ArrayElement.h"
#include "base/util/WString.h"
#include "base/util/StringBuffer.h"
#include "base/util/ArrayList.h"
#include "base/globalsdef.h"
#include "base/winadapter.h"
#include "base/fscapi.h"
#include "base/adapter/PlatformAdapter.h"
#include "pim/mail/MessagingStatus.h"
#include "pim/ClientSettings.h"

USE_NAMESPACE

MessagingStatus* MessagingStatus::pinstance         = NULL;
const wchar_t* MessagingStatus::cacheSemaphoreName  = L"Global\\MessagingStatusCache";
const unsigned int MessagingStatus::semaphoreTimeOut = 500; // half a second

MessagingStatus* MessagingStatus::getInstance()
{
    if (pinstance == NULL) {
        pinstance = new MessagingStatus();
    }

    return pinstance;
}

void MessagingStatus::disposeInstance()
{
    if (pinstance) {
        delete pinstance;
        pinstance = NULL;
    }
}

MessagingStatus::MessagingStatus()
{
    // set up cache and temporary cache paths
    string programPath = getRegConfig()->getPath();
    StringBuffer cachePath = programPath.c_str();
    cacheName = cachePath + "/" + MESSAGE_STATUS_CACHE_NAME;

    cacheSemaphore = CreateSemaphore(NULL, 1, 1, cacheSemaphoreName);

    if (cacheSemaphore == NULL) {
        LOG.error("error creating cache semaphore: %d\n", GetLastError());
    } 

    readCache();
}

MessagingStatus::~MessagingStatus()
{
    storeCache();

    // release global semaphore
    if (cacheSemaphore) {
        CloseHandle(cacheSemaphore);
    }
}

bool MessagingStatus::checkCacheFile() const
{
    return fileExists(cacheName.c_str());
}

const ItemData* MessagingStatus::getItem(const wstring& itemId)
{
    int cacheSize = 0;
    ItemData* itemData = NULL;

    if (readCache() == false) {
        LOG.error("error reading message status cache");
    } else {
        cacheSize = itemStatusCache.size();
    
        for (int i = 0; i < cacheSize; i++) {
            itemData = static_cast<ItemData *>(itemStatusCache[i]);
            if (itemData) {
                const wstring& id = itemData->getId();
                if (id == itemId) {
                    break;
                }
            }
        }
    }

    itemStatusCache.clear();

    return itemData;
}


int MessagingStatus::findItem(const wstring& itemId) const
{
    int pos = -1;
    int cacheSize = itemStatusCache.size();

    for (int i = 0; i < cacheSize; i++) {
        ItemData* itemData = static_cast<ItemData *>(itemStatusCache[i]);
        if (itemData) {
            const wstring& id = itemData->getId();
            if (id == itemId) {
                return i;
            }
        }
    }

    return pos;
}

void MessagingStatus::getItemsByStatus(ArrayList& cache, MapiItemStatus status)
{
    int cacheSize = 0;
    
    itemStatusCache.clear();

    if (readCache() == false) {
        LOG.error("error reading message status cache");

        return;
    }

    cacheSize = itemStatusCache.size();

    for (int i = 0; i < cacheSize; i++) {
        ItemData* itemData = static_cast<ItemData *>(itemStatusCache[i]);
        if (itemData) {
            MapiItemStatus itemStatus = itemData->getStatus();
            if (status == itemStatus) {
                cache.add(static_cast<ItemData>(*itemData));
            }
        }
    }

    itemStatusCache.clear();
}

bool MessagingStatus::addItem(const ItemData& item)
{
    const wstring& id = item.getId();

    if (id.empty()) {
        LOG.error("invalid entry: can't store on messaging status cache");

        return false;
    }

    if (checkCacheLock()) {
        LOG.info("message status cache locked");

        return true;
    } 

    // fill up temporary cache with stored items
    if (readCache() == false) {
        LOG.error("error reading message status cache");

        return false;
    }

    // check if item is already in cache to avoid duplicate entries on add
    if (findItem(id) >= 0) {
        LOG.error("trying to add duplicate item in message status cache");
        return false;
    }

    // update cache
    itemStatusCache.add(static_cast<ItemData>(item));

    // store cache on filesystem
    if (storeCache() == false) {
        itemStatusCache.clear();

        return false;
    }

    // cleanup temporary cache
    itemStatusCache.clear();
    
    return true;
}

bool MessagingStatus::updateItem(const ItemData& item)
{
    int pos = 0;
    bool skipItem = false;
   
    if (checkCacheLock()) {
        return true;
    }

    if (readCache() == false) {
        LOG.error("error reading message status cache");

        return false;
    }

    const wstring& id = item.getId();

    // if item is present in cache remove it
    if ((pos = findItem(id)) >= 0) {
        MapiItemStatus newStatus = item.getStatus();
        ItemData* oldItem = static_cast<ItemData*>(itemStatusCache.get(pos));

        if (oldItem == NULL) {
            LOG.error("error getting item from message status cache");
            itemStatusCache.clear();

            return false;
        }

        MapiItemStatus oldStatus = oldItem->getStatus();
        
        // remove old element element in memory cache 
        itemStatusCache.removeElementAt(pos);    

        if ((oldStatus == item_status_new) && (newStatus == item_status_deleted)) {
            // item was added and then removed: discard it
            skipItem = true;  
        }
    }

    // add updated item
    if (skipItem == false) {
        itemStatusCache.add(static_cast<ItemData>(item));
        // update cache on store
        if (storeCache() == false) {
            itemStatusCache.clear();
            return false;
        }
    }

    itemStatusCache.clear();

    return true;
}

bool MessagingStatus::removeCacheItem(const wstring& id)
{
    int pos = 0;

    if (readCache() == false) {
        LOG.error("error reading message status cache");

        return false;
    }

    if ((pos = findItem(id)) < 0) {
        LOG.error("item not found in cache");
        return false;
    }

    // remove element from cache
    itemStatusCache.removeElementAt(pos);

    if (storeCache() == false) {
        itemStatusCache.clear();
        return false;
    }

    itemStatusCache.clear();

    return true;
}


const ArrayList& MessagingStatus::getNewItems()
{
    // clean up temporary cache
    newItemsCache.clear();
    getItemsByStatus(newItemsCache, item_status_new);

    return newItemsCache;
}

const ArrayList& MessagingStatus::getUpdatedItems()
{
    // clean up temporary cache
    updatedItemsCache.clear();
    getItemsByStatus(updatedItemsCache, item_status_updated);

    return updatedItemsCache;    
}

const ArrayList& MessagingStatus::getDeletedItems()
{
    // clean up temporary cache
    deletedItemsCache.clear();
    getItemsByStatus(deletedItemsCache, item_status_deleted);

    return deletedItemsCache;
}

bool MessagingStatus::commit()
{
    // persist memory cache
    if (storeCache() == false) {
        LOG.error("error writing cache to file system");

        return false;
    }

    // clean up cache
    itemStatusCache.clear();

    return true;
}


bool MessagingStatus::refreshCache()
{
    return readCache();
}

bool MessagingStatus::readCache()
{
    char* cacheStr = NULL;
    size_t cacheBufferSize = 0;
    DWORD dwWaitRes; 
    
    if (fileExists(cacheName.c_str()) == false) {
        LOG.debug("message status cache file not yet created");

        return true;
    }

    // try to get access to global semaphore 
    dwWaitRes = WaitForSingleObject(cacheSemaphore, semaphoreTimeOut);           

    switch (dwWaitRes) {
        case WAIT_OBJECT_0: {
            bool ret = readFile(cacheName.c_str(), &cacheStr, &cacheBufferSize, false);

            if (!ReleaseSemaphore(cacheSemaphore, 1, NULL)) {
                LOG.error("error releasing message status cache semaphore");
            }

            if (ret == false) {
                LOG.error("error reading message status cache");
        
                return false;
            }

            break;
        }
        
        case WAIT_TIMEOUT:
            LOG.error("can't access message status cache in locked state");
            return false;

        default:
            LOG.error("unexpected error from WaitForSingleObject");
            return false;
    }

    if ((cacheStr == NULL) && (cacheBufferSize == 0)) {
        LOG.debug("no data found in message status cache");

        return true;
    }
    
    StringBuffer cacheBuf(cacheStr);
    ArrayList cacheLines;
    // split cache by lines;
    cacheBuf.split(cacheLines, "\n");

    const StringBuffer* line;

    // update memory cache with persisted  data from file system
    for (line=(StringBuffer *)cacheLines.front(); line; line=(StringBuffer *)cacheLines.next() ) {
        ArrayList lineItem;

        // item data fields
        MapiItemStatus status;
        StringBuffer parent_role;
        unsigned long type = 0;
        bool read_flag = false;
        bool incomplete_flag = false;

        // silently discard empty lines
        if (line->empty()) {
            continue;
        }

        // split each cache line in ItemData member
        line->split(lineItem, "-");
        
        int itemSize = lineItem.size();
        // check validity of cache line
        if (itemSize != messageStatusCacheItemsPerLine) {
            LOG.error("invalid item in message status cache line: %d", itemSize);
            continue;
        }

        // get object ID
        StringBuffer* id_str = static_cast<StringBuffer *>(lineItem[0]);
        if (id_str->empty()) {
            LOG.error("invalid item id in message status cache line");
            continue;
        }

        // get object type
        StringBuffer* type_str = static_cast<StringBuffer*>(lineItem[1]);
        if (*type_str == MAPI_MESSAGE_ID) {
            type = MAPI_MESSAGE;
        } else if (*type_str == MAPI_FOLDER_ID) {
            type = MAPI_FOLDER;
        } else if (*type_str == MAPI_ACCOUNT_ID) {
            type = MAPI_STORE;
        } else {
            LOG.error("invalid item type in message status cache line");
            continue;
        }

        // get object status
        StringBuffer* status_str = static_cast<StringBuffer*>(lineItem[2]);
        if (*status_str == STATUS_NEW_ID) {
            status = item_status_new;
        } else if (*status_str == STATUS_UPDATED_ID) {
            status = item_status_updated;
        } else if (*status_str == STATUS_DELETED_ID) {
            status = item_status_deleted;
        } else {
            LOG.error("invalid item status in message status cache line");
            continue;
        }

        // get object parent ID
        StringBuffer* parent_id_str = static_cast<StringBuffer*>(lineItem[3]);
        if (parent_id_str->empty()) {
            LOG.error("invalid item parent id in message cache line");
            continue;
        }
        
        // get object parent role
        StringBuffer* parent_role_str = static_cast<StringBuffer*>(lineItem[4]);
        if (*parent_role_str == PARENT_ROLE_INBOX) {
            parent_role = ROLE_INBOX;
        } else if (*parent_role_str == PARENT_ROLE_OUTBOX) {
            parent_role = ROLE_OUTBOX;
        } else {
            LOG.error("invalid item parent role in message cache line");
            continue;
        }

        // get status flags
        StringBuffer* itemField;
        itemField = static_cast<StringBuffer *>(lineItem[5]);

        if (*itemField == "true") {
            read_flag = true;
        }
        
        itemField = static_cast<StringBuffer *>(lineItem[6]);

        if (*itemField == "true") {
            incomplete_flag = true;
        }

        const WCHAR* id = toWideChar(*id_str);
        const WCHAR* parent_id = toWideChar(*parent_id_str);

        LOG.debug("message status cache: adding read item to cache");
        ItemData itemData(id, parent_id, type, status, parent_role, read_flag, incomplete_flag);
        itemStatusCache.add(itemData);

        delete [] id;
        delete [] parent_id;
    }

    return true;
}

bool MessagingStatus::storeCache()
{
    int cacheSize = itemStatusCache.size();
    StringBuffer cacheString;

    // format each element of the cache as a string
    // and store to file
    for (int i = 0; i < cacheSize; i++) {
        ItemData* itemData = static_cast<ItemData *>(itemStatusCache[i]);
        
        if (itemData) {
            StringBuffer cacheLine;

            if (formatCacheLine(itemData, cacheLine) == false) {
                LOG.error("invalid mapi item entry in message status cache");
            } else {
                if (cacheLine.empty() == false) {
                    cacheString.append(cacheLine);
                }
            }
        }
    }

    if (cacheString.empty()) {
        // no entry to save: just remove cache file
        if (removeCache() == false) {
            LOG.error("error removing message status cache file");

            return false;
        }
    } else { // store data on file
        DWORD dwWaitRes = WaitForSingleObject(cacheSemaphore, semaphoreTimeOut);           

        // try to get access to global semaphore 
        switch (dwWaitRes) {
            case WAIT_OBJECT_0: {
                bool ret = saveFile(cacheName.c_str(), cacheString.c_str(), 
                                    cacheString.length(), false);

                if (!ReleaseSemaphore(cacheSemaphore, 1, NULL)) {
                    LOG.error("error releasing message status cache semaphore");
                }

                if (ret == false) {
                    LOG.error("error storing message status cache on file system");  
                    return false;
                }
                break;
            }
        
            case WAIT_TIMEOUT:
                LOG.error("can't access message status cache in locked state");
                return false;

            default:
                LOG.error("unexpected error from WaitForSingleObject");
                return false;
        }
    }

    return true;
}

bool MessagingStatus::resetCache()
{
    // remove memory cache
    itemStatusCache.clear();
    // delete store
    if (removeCache() == false) {
        return false;
    }
    
    return true;
}

bool MessagingStatus::removeCache()
{
    // set up cache and temporary cache paths
    string programPath = getRegConfig()->getPath();
    StringBuffer cachePath = programPath.c_str();
    
    itemStatusCache.clear();

    if (removeFileInDir(cachePath.c_str(), MESSAGE_STATUS_CACHE_NAME) == false) {
        LOG.error("error removing message status cache file");
        return false;
    }

    return true;
}

bool MessagingStatus::formatCacheLine(const ItemData* item, StringBuffer& cacheLine)
{
    // item data to be formatted in cache
    wstring id;
    wstring parent_id;
    unsigned long type = 0;
    MapiItemStatus status;
    StringBuffer parent_role;
    bool read_flag(false);
    bool incomplete_flag(false);

    // item data as string values
    const char* item_id_str = NULL;
    const char* item_parent_id_str = NULL;
    const char* item_type_str = NULL;
    const char* item_status_str = NULL;
    const char* item_parent_role = NULL;
    const char* cacheLineFmt = "%s-%s-%s-%s-%s-%s-%s\n";

    id = item->getId();

    if (id.empty()) {
        LOG.error("invalid id for item in message status cache");
        return false;
    }

    parent_id = item->getParentId();
     
    type = item->getType();

    switch (type) {
        case MAPI_STORE:
            item_type_str = MAPI_ACCOUNT_ID;
            break;
        case MAPI_FOLDER:
            item_type_str = MAPI_FOLDER_ID;
            break;
        case MAPI_MESSAGE:
            item_type_str = MAPI_MESSAGE_ID;
            break;
        default:
            LOG.error("invalid object type for item in cache");
            return false;
    }

    status = item->getStatus();

    switch (status) {
        case item_status_new:
            item_status_str = STATUS_NEW_ID;
            break;
        case item_status_updated:
            item_status_str = STATUS_UPDATED_ID;
            break;
        case item_status_deleted:
            item_status_str = STATUS_DELETED_ID;
            break;
        default:
            LOG.error("invalid item status in cache");
            return false;
    }

    item_id_str = toMultibyte(id.c_str());

    if (parent_id.empty() == false) {
        item_parent_id_str = toMultibyte(parent_id.c_str());
    }

    parent_role = item->getParentRole();

    if (parent_role == ROLE_INBOX) {
        item_parent_role = PARENT_ROLE_INBOX;
    } else if (parent_role == ROLE_OUTBOX) {
        item_parent_role = PARENT_ROLE_OUTBOX;
    } 

    read_flag = item->getReadFlag();
    incomplete_flag = item->getIncompleteFlag();

    cacheLine.sprintf(cacheLineFmt, item_id_str, item_type_str, item_status_str, 
        item_parent_id_str ? item_parent_id_str : "", 
        item_parent_role ? item_parent_role : "", 
        read_flag ? "true" : "false", 
        incomplete_flag ? "true" : "false");

    delete [] item_id_str;
    item_id_str = NULL;

    if (item_parent_id_str) {
        delete [] item_parent_id_str;
        item_parent_id_str = NULL;
    }

    return true;
}

bool MessagingStatus::lockCache()
{
    bool locked = false;

    if (cacheSemaphore == NULL) {
        LOG.error("messaging status lock not initialized");

        return false;
    }

    // try to get access to global semaphore 
    DWORD dwWaitRes = WaitForSingleObject(cacheSemaphore, semaphoreTimeOut);           

    switch (dwWaitRes) {
        case WAIT_OBJECT_0: {
            locked = true;
            break;
        }

        case WAIT_TIMEOUT: {
            locked = false;
            break;
        }

        default: {
            LOG.error("WaitForSingleObject unexpected error: %d", dwWaitRes);
            locked = false;
            break;
        }
    }

    return locked;
}

bool MessagingStatus::unlockCache()
{
    bool unlocked = false;

    if (cacheSemaphore == NULL) {
        LOG.error("messaging status lock not initialized");

        return unlocked;
    }

    if (!ReleaseSemaphore(cacheSemaphore, 1, NULL)) {
        LOG.error("error releasing message status cache semaphore");
    } else {
        unlocked = true;
    }

    return unlocked;
}


bool MessagingStatus::checkCacheLock(unsigned long timeout)
{
    bool locked = true;

    if (cacheSemaphore == NULL) {
        LOG.error("messaging status lock not initialized");

        return false;
    }

    // try to get access to global semaphore 
    DWORD dwWaitRes = WaitForSingleObject(cacheSemaphore, timeout);           

    switch (dwWaitRes) {
        case WAIT_OBJECT_0: {
            if (!ReleaseSemaphore(cacheSemaphore, 1, NULL)) {
                LOG.error("error releasing message status cache semaphore");
            }
            
            locked = false;
            break;
        }

        case WAIT_TIMEOUT:
            locked = true;
            break;

        default:
            LOG.error("WaitForSingleObject unexpected error: %d", dwWaitRes);
            locked = true;
            break;
    }

    return locked;
}

// exported methods for Unit Testing

MessagingStatus* getMessagingStatus()
{
    return MessagingStatus::getInstance();
}

bool MessagingStatus::addMessagingStatusItem(const ItemData& item)              { return addItem(item); }
bool MessagingStatus::updateMessagingStatusItem(const ItemData& item)           { return updateItem(item); }
const ArrayList& MessagingStatus::getMessagingStatusNewItems()                  { return getNewItems(); }
const ArrayList& MessagingStatus::getMessagingStatusUpdatedItems()              { return getUpdatedItems(); }
const ArrayList& MessagingStatus::getMessagingStatusDeletedItems()              { return getDeletedItems(); }
const ItemData* MessagingStatus::getMessagingStatusItem(const wstring& itemId)  { return getItem(itemId); }
bool MessagingStatus::lockMessagingStatusCache()                                { return lockCache();  }
bool MessagingStatus::unlockMessagingStatusCache()                              { return unlockCache(); }
bool MessagingStatus::checkMessagingStatusCache() const                         { return checkCacheFile(); } 
