/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2009 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include <cppunit/extensions/TestFactoryRegistry.h>
#include <cppunit/extensions/HelperMacros.h>
#include "base/globalsdef.h"
#include "base/adapter/PlatformAdapter.h"
#include "customization.h"
#include "base/fscapi.h"
#include "base/util/utils.h"
#include "pim/mail/MessagingStatus.h"


USE_NAMESPACE

class MessagingStatusTest : public CppUnit::TestFixture {
private:
    CPPUNIT_TEST_SUITE(MessagingStatusTest);

    CPPUNIT_TEST(testMessagingStatusAddItem);
    CPPUNIT_TEST(testMessagingStatusUpdateItem)
    CPPUNIT_TEST(testMessagingStatusDeleteItem);
    CPPUNIT_TEST(testMessagingStatusGetItem);
    CPPUNIT_TEST(testMessagingStatusCacheLocked);
    CPPUNIT_TEST(testMessagingStatusReadCache);
    
    CPPUNIT_TEST_SUITE_END();

    MessagingStatus* msgStatus;

    // add mail type items to the cache
    bool addMessageItem(const wstring& _id, const wstring& _parent_id) {
        wstring id(_id);
        wstring parent_id(_parent_id);
        MapiItemStatus status = item_status_new;
        unsigned long type = MAPI_MESSAGE;

        ItemData item(id, parent_id, type, status);
       
        // try add the item
        bool ret = msgStatus->addMessagingStatusItem(item);

        return ret;
    }

    bool updateMessageItem(const wstring& _id, const wstring& _parent_id, 
            MapiItemStatus _status, bool _read=false, bool _incomplete=false) {
        wstring id(_id);
        wstring parent_id(_parent_id);
        unsigned long type = MAPI_MESSAGE;
        bool read_flag=_read;
        bool incomplete_flag =_incomplete;

        ItemData item(id, parent_id, type, _status, read_flag, incomplete_flag);
       
        // try update the item
        bool ret = msgStatus->updateMessagingStatusItem(item);

        return ret;
    }

public:

    void setUp() {
        PlatformAdapter::init(ROOT_CONTEXT);
        msgStatus = getMessagingStatus();
        CPPUNIT_ASSERT_MESSAGE("error getting message status cache instance", msgStatus != NULL);

        // clean up cache to avoid reading invalid data
        msgStatus->resetCache();
    }

    void tearDown() {
        msgStatus->resetCache();
    }


    void testMessagingStatusAddItem() {
        bool ret = true;
        wstring id(L"ID");
        wstring parent_id(L"PARENT_ID");
        MapiItemStatus status = item_status_new;
        unsigned long type = MAPI_MESSAGE;

        ItemData item(id, parent_id, type, status);
        
        ret = msgStatus->addMessagingStatusItem(item);
        CPPUNIT_ASSERT(ret == true);

        msgStatus->resetCache();
    }

    void testMessagingStatusUpdateItem() {
        bool ret = true;
        wstring id(L"ID");
        wstring parent_id(L"PARENT_ID");
        MapiItemStatus status = item_status_updated;
        unsigned long type = MAPI_MESSAGE;
        
        ItemData item(id, parent_id, type, status);
        
        ret = msgStatus->addMessagingStatusItem(item);
        CPPUNIT_ASSERT_MESSAGE("failed to add item in cache", ret == true);

        bool read = true;
        status = item_status_updated;
        item = ItemData(id, parent_id, type, status, read);
        ret = msgStatus->updateMessagingStatusItem(item);
        CPPUNIT_ASSERT(ret == true);

        msgStatus->resetCache();
    }

    void testMessagingStatusDeleteItem() 
    {
        bool ret = true;
        
        ret = addMessageItem(L"ID", L"PARENT_ID");
        CPPUNIT_ASSERT_MESSAGE("failed to add item in cache", ret == true);

        ret = updateMessageItem(L"ID", L"PARENT_ID", item_status_deleted, true);
        CPPUNIT_ASSERT_MESSAGE("failed to update item in cache", ret == true);

        msgStatus->resetCache();
    }

    void testMessagingStatusGetItem() 
    {
        bool ret = true;

        // add two items
        ret = addMessageItem(L"ID1", L"PARENT_ID");
        CPPUNIT_ASSERT_MESSAGE("failed to add item in cache", ret == true);

        ret = addMessageItem(L"ID2", L"PARENT_ID");
        CPPUNIT_ASSERT_MESSAGE("failed to add item in cache", ret == true);

        ArrayList newItems = msgStatus->getMessagingStatusNewItems();
        int cacheSize = newItems.size();
        CPPUNIT_ASSERT_MESSAGE("new item missing in cache", cacheSize == 2);

        ret = updateMessageItem(L"ID1", L"PARENT_ID", item_status_updated, true); 
        CPPUNIT_ASSERT_MESSAGE("error updating item in cache", ret == true);

        ArrayList updatedItems = msgStatus->getMessagingStatusUpdatedItems();
        cacheSize = updatedItems.size();
        CPPUNIT_ASSERT_MESSAGE("error getting updated items from cache", cacheSize == 1);

        newItems = msgStatus->getMessagingStatusNewItems();
        cacheSize = newItems.size();
        CPPUNIT_ASSERT_MESSAGE("error getting new items from cache", cacheSize == 1);

        msgStatus->resetCache();
    }

    void testMessagingStatusCacheLocked() 
    {
        bool ret = true;

        // lock cache
        msgStatus->lock();
        ret = addMessageItem(L"ID", L"PARENT_ID");
        CPPUNIT_ASSERT_MESSAGE("failed to add item in cache", ret == true);
        
        // check if cache was updated
        ArrayList newItems = msgStatus->getMessagingStatusNewItems();
        int cacheSize = newItems.size();

        CPPUNIT_ASSERT_MESSAGE("item added when cache was in locked state", cacheSize == 0);

        // unlock 
        msgStatus->unlock();

        // insert again an item
        ret = addMessageItem(L"ID", L"PARENT_ID");
        CPPUNIT_ASSERT_MESSAGE("failed to add item in cache", ret == true);
        
        // check if cache was updated
        newItems = msgStatus->getMessagingStatusNewItems();
        cacheSize = newItems.size();

        CPPUNIT_ASSERT_MESSAGE("item missing in cache", cacheSize == 1);

        msgStatus->resetCache();
    }

    void testMessagingStatusReadCache() {
        bool ret = false;

        // add some item to cache
        ret = addMessageItem(L"ID1", L"PARENT_ID");
        CPPUNIT_ASSERT_MESSAGE("failed to add item in cache", ret == true);

        ret = addMessageItem(L"ID2", L"PARENT_ID");
        CPPUNIT_ASSERT_MESSAGE("failed to add item in cache", ret == true);

        ret = addMessageItem(L"ID3", L"PARENT_ID");
        CPPUNIT_ASSERT_MESSAGE("failed to add item in cache", ret == true);

        // deleting MessagingStatus singleton will force cache storing
        msgStatus->disposeInstance();

        // by getting a fresh instance of MessagingStatus cache will be read
        msgStatus = getMessagingStatus();
        CPPUNIT_ASSERT_MESSAGE("failed to create MessagingStatus instance", msgStatus != NULL);

        // read added elements from memory cache
        ArrayList newItems = msgStatus->getMessagingStatusNewItems();
        int cacheSize = newItems.size();
        CPPUNIT_ASSERT_MESSAGE("items not found cache", cacheSize == 3);

        msgStatus->resetCache();
    }
};

//CPPUNIT_TEST_SUITE_REGISTRATION( MessagingStatusTest );
