﻿/////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2007, Microsoft All Rights Reserved.
//
// THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
/////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.DeviceEmulatorManager.Interop;
using System.Runtime.InteropServices;

namespace WindowsMobileDeveloperSamples.DeviceEmulatorManagerAutomationWrapper
{
    /// <summary>
    /// Represents a specific device emulator. Each device emulator is
    /// grouped within an SDK
    /// </summary>
    public class DeviceEmulator
    {
        internal DeviceEmulator(DeviceEmulatorSdk sdk)
        {
            _nativeDeviceEmulatorManagerVMID = sdk.NativeEnumVMIDs.GetVMID();
        }

        // Underlying Native COM Interface Reference
        private IDeviceEmulatorManagerVMID _nativeDeviceEmulatorManagerVMID;
        internal IDeviceEmulatorManagerVMID NativeDeviceEmulatorManagerVMID
        { get { return _nativeDeviceEmulatorManagerVMID; } }

        /// <summary>
        /// Shows the Device Emulator window
        /// <remarks>
        /// See the IDeviceEmulatorManagerVMID::BringToFront documentation for more details
        /// </remarks>
        /// </summary>
        public void BringToFront()
        {
            _nativeDeviceEmulatorManagerVMID.BringToFront();
        }

        /// <summary>
        /// Clears the Device Emulator saved state file (.dess) associated with this device. 
        /// The next time that you start, the emulator will start from its ROM image instead 
        /// of its saved state file.
        /// <remarks>
        /// See the IDeviceEmulatorManagerVMID::ClearSaveState documentation for more details
        /// </remarks>
        /// </summary>
        public void ClearSaveState()
        {
            _nativeDeviceEmulatorManagerVMID.ClearSaveState();
        }

        /// <summary>
        /// Starts the Device Emulator if it is not already started and connects to it.
        /// <remarks>
        /// See the IDeviceEmulatorManagerVMID::Connect documentation for more details
        /// </remarks>
        /// </summary>
        public void Connect()
        {
            _nativeDeviceEmulatorManagerVMID.Connect();
        }

        /// <summary>
        /// Starts the Device Emulator if it is not already started and connects to it.
        /// <remarks>
        /// See the IDeviceEmulatorManagerVMID::Cradle documentation for more details
        /// </remarks>
        /// </summary>
        public void Cradle()
        {
            _nativeDeviceEmulatorManagerVMID.Cradle();
        }

        /// <summary>
        /// Gets the configuration of the emulator in Device Emulator Configuration XML format.
        /// <remarks>
        /// See the IDeviceEmulatorManagerVMID::GetConfiguration documentation for more details
        /// </remarks>
        /// </summary>
        /// <returns>A string that contains the emulator's configuration</returns>
        public string GetConfiguration()
        {
            return _nativeDeviceEmulatorManagerVMID.GetConfiguration();
        }
        
        /// <summary>
        /// Gets the name of the emulator, for example, Pocket PC 2003 SE Emulator.
        /// <remarks>
        /// See the IDeviceEmulatorManagerVMID::get_Name documentation for more details
        /// </remarks>
        /// </summary>
        public string Name { get { return _nativeDeviceEmulatorManagerVMID.get_Name(); } }

        /// <summary>
        /// Performs a soft-reset or hard-reset of the emulator as specified by the softReset parameter.
        /// </summary>
        /// <param name="softReset">Whether to perform a soft-reset or hard-reset. A value true indicates a soft-reset.</param>
        public void Reset(bool softReset)
        {
            int intSoftReset = softReset ? 1 : 0;
            _nativeDeviceEmulatorManagerVMID.Reset(intSoftReset);
        }
        /// <summary>
        /// Sets the configuration of the emulator given a correctly formatted Device 
        /// Emulator Configuration XML string.
        /// <remarks>
        /// See the IDeviceEmulatorManagerVMID::SetConfiguration documentation for more details
        /// </remarks>
        /// </summary>
        /// <param name="configurationXmlAsString"></param>
        public void SetConfiguration(string configurationXmlAsString)
        {
            _nativeDeviceEmulatorManagerVMID.SetConfiguration(configurationXmlAsString);
        }

        /// <summary>
        /// Shuts down the emulator and, optionally, saves the emulator's state.
        /// <remarks>
        /// See the IDeviceEmulatorManagerVMID::Shutdown documentation for more details
        /// </remarks>
        /// </summary>
        /// <param name="saveState">Indicates whether to save emulator state before exiting. 
        /// If true, the device emulator saves the emulator's current state in a .dess 
        /// file before exiting.</param>
        public void Shutdown(bool saveState)
        {
            int intSaveState = saveState ? 1 : 0;
            _nativeDeviceEmulatorManagerVMID.Shutdown(intSaveState);
        }

        /// <summary>
        /// Gets the current state of the emulator.
        /// <remarks>
        /// See the IDeviceEmulatorManagerVMID::get_State documentation for more details
        /// </remarks>
        /// </summary>
        public EMULATOR_STATE State { get { return _nativeDeviceEmulatorManagerVMID.get_State(); } }

        /// <summary>
        /// Uncradles the emulator.
        /// <remarks>
        /// See the IDeviceEmulatorManagerVMID::UnCradle documentation for more details
        /// </remarks>
        /// </summary>
        public void UnCradle()
        {
            _nativeDeviceEmulatorManagerVMID.UnCradle();
        }

        /// <summary>
        /// Gets the virtual machine identifier (VMID) that uniquely identifies the emulator.
        /// <remarks>
        /// See the IDeviceEmulatorManagerVMID::get_VMID documentation for more details
        /// </remarks>
        /// </summary>
        public string VMID { get { return _nativeDeviceEmulatorManagerVMID.get_VMID(); } }

        public override string ToString()
        {
            return Name;
        }
    }
}
