﻿/////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2007, Microsoft All Rights Reserved.
//
// THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
/////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.DeviceEmulatorManager.Interop;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;

namespace WindowsMobileDeveloperSamples.DeviceEmulatorManagerAutomationWrapper
{
    /// <summary>
    /// Represents the Device Emulator Manager 
    /// </summary>
    public class DeviceEmulatorManagerEx : IDisposable 
    {
        const int END_OF_DATA = -2147024637;

        public DeviceEmulatorManagerEx()
        {
            _nativeDeviceEmulatorManager = new DeviceEmulatorManagerClass();
        }

        public void Dispose()
        {
            ShutdownEventListener();
        }

        // Underlying Native COM Interface Reference
        private IDeviceEmulatorManager _nativeDeviceEmulatorManager;
        public IDeviceEmulatorManager NativeDeviceEmulatorManager { get { return _nativeDeviceEmulatorManager; } }

        /// <summary>
        /// Causes DeviceEmulatorManagerEx instance to re-read the category, sdk, emulator hierarchy
        /// </summary>
        public void Refresh()
        {
            _categories = null; // Clear existing cache of information
            _nativeDeviceEmulatorManager.Refresh();
        }

        /// <summary>
        /// Shows or hides the Device Emulator Manager window
        /// </summary>
        /// <param name="show">
        /// Indicates whether to show or hide the Device Emulator Manager window. 
        /// Use true to show the window, and false to hide the window.
        /// </param>
        public void ShowManagerUI(bool show)
        {
            int showAsInteger = show ? 1 : 0;
            _nativeDeviceEmulatorManager.ShowManagerUI(showAsInteger);
        }

        /// <summary>
        /// Locates a device emulator using the device emulator's name or unique VMID
        /// </summary>
        /// <param name="nameOrVMID">Name or unique VMID of Emulator to find</param>
        /// <returns>A reference to an Emulator with the specified name or VMID;
        /// null if not found.</returns>
        public DeviceEmulator FindDeviceEmulator(string emulatorNameOrVMID)
        {
            foreach (DeviceEmulatorCategory category in Categories)
                foreach (DeviceEmulatorSdk sdk in category.Sdks)
                    foreach (DeviceEmulator deviceEmulator in sdk.DeviceEmulators)
                        if (string.Compare(deviceEmulator.Name, emulatorNameOrVMID, true) == 0 ||
                            string.Compare(deviceEmulator.VMID, emulatorNameOrVMID, true) == 0)
                            return deviceEmulator;
            return null;
        }
        
        private DeviceEmulatorCategoryCollection _categories = null;
        /// <summary>
        /// The list of Device Emulator Manager Categories
        /// </summary>
        public DeviceEmulatorCategoryCollection Categories
        {
            get
            {
                if (_categories == null)
                {
                    List<DeviceEmulatorCategory> categoryList = BuildCategoryList();
                    _categories = new DeviceEmulatorCategoryCollection(categoryList);
                }
                return _categories;
            }
        }

        private List<DeviceEmulatorCategory> BuildCategoryList()
        {
            List<DeviceEmulatorCategory> categoryList = new List<DeviceEmulatorCategory>();
            DeviceEmulatorCategory category;

            try
            {
                _nativeDeviceEmulatorManager.Reset();
                while (true)
                {
                    category = new DeviceEmulatorCategory(this);
                    categoryList.Add(category);
                    _nativeDeviceEmulatorManager.MoveNext();
                }
            }
            catch (COMException ex)
            {
                // API signals end-of-data with a No More Data HResult.
                if (ex.ErrorCode != END_OF_DATA)
                    throw ex;
            }

            return categoryList;
        }

        MethodInvoker _refreshRequired;
        /// <summary>
        /// Signals when the DeviceEmulatorManagerEx the hierarchy of categories, sdks, and emuulators
        /// must be refreshed because of changes in the Datastore, My Device Emulators, or All Device Emulators catories.
        /// <remarks>Use caution when interacting with the user interface from a RefreshRequired event handler.
        /// The RefreshRequired event is signaled using a non-UI thread and therefore the associated event
        /// handlers cannot interact directly with the UI; instead they must use the 
        /// Control.Invoke or Control.BeginInvoke methods.<BR/>
        /// See the IDeviceEmulatorManager::RegisterRefreshEvent documentation for more details.</remarks>
        /// </summary>
        public event MethodInvoker RefreshRequired
        {
            add
            {
                if (_refreshRequired == null)
                    SetupEventListener(); // first event handler so start listening
                _refreshRequired += value;
            }
            remove
            {
                _refreshRequired -= value;
                if (_refreshRequired == null)
                    ShutdownEventListener(); // No more event handlers so clear listening overhead
            }
        }

        #region RefreshRequired Event implementation details

        // Name of event that DeviceEmulatorManager API will signal when data needs to be refreshed
        const string _refreshEventNameBase = "DeviceEmulatorManagerEx.RefreshEvent.";
        const string _refreshThreadName = "RefreshEventMonitor.";
        string _refreshEventName;
        EventWaitHandle _refreshEvent;
        Thread _refreshEventThread;
        volatile bool _shutDownRequested = false;

        private void SetupEventListener()
        {
            // Append GUID to event name to assure that doesn't conflict with any other
            _refreshEventName = _refreshEventNameBase + Guid.NewGuid().ToString();
            _refreshEvent = new EventWaitHandle(false, EventResetMode.AutoReset, _refreshEventName);
            _shutDownRequested = false;
            _refreshEventThread = new Thread(RefreshEventMonitor);
            _refreshEventThread.Name = _refreshThreadName;
            _refreshEventThread.Start();
            _nativeDeviceEmulatorManager.RegisterRefreshEvent(_refreshEventName);
        }

        const int _shutDownEventTimeoutMilliseconds = 1000; // 1 second
        private void ShutdownEventListener()
        {
            if (_refreshEventThread != null)
            {
                _nativeDeviceEmulatorManager.UnRegisterRefreshEvent();
                _shutDownRequested = true;
                _refreshEvent.Set();
                bool shutdownSuccessful = _refreshEventThread.Join(_shutDownEventTimeoutMilliseconds);
                if (!shutdownSuccessful)
                    _refreshEventThread.Abort();
                _refreshEvent.Close();
                _refreshEvent = null;
                _refreshEventThread = null;
            }
        }

        void RefreshEventMonitor()
        {
            while (!_shutDownRequested)
            {
                bool signaled = _refreshEvent.WaitOne();
                if (signaled && !_shutDownRequested)
                    FireEvents();
            }
        }

        private void FireEvents()
        {
            if (_refreshRequired != null)
                _refreshRequired();
        }

        #endregion

    }
}
